import numpy as np
from ase.calculators.lj import LennardJones

from kim_tools import get_isolated_energy_per_atom

# from lj_fail_no_neighbors import LennardJonesFailNoNeighbors


def test_get_isolated_energy_per_atom():
    for model in [
        LennardJones(),
        "LennardJones612_UniversalShifted__MO_959249795837_003",
        "Sim_LAMMPS_LJcut_AkersonElliott_Alchemy_PbAu",
    ]:
        for species in ["Pb", "Au"]:
            assert np.isclose(
                get_isolated_energy_per_atom(model=model, symbol=species),
                0,
            )
    # assert np.isclose(
    #     get_isolated_energy_per_atom(LennardJonesFailNoNeighbors(), "H"),
    #    0,
    # )



import logging
import math
import os
import shutil
from typing import Dict, List, Tuple, Union

import numpy as np
import numpy.typing as npt

from .core import DATA_DIR, _check_space_group

logger = logging.getLogger(__name__)
logging.basicConfig(filename="kim-tools.log", level=logging.INFO, force=True)


def voigt_elast_class(sgnum: Union[int, str]) -> str:
    """
    Get the name of the class of the structure of the elasticity tensor, out of
    the following possibilities:
    "cubic", "hexagonal", "trigonal_3bar_m_2nd_pos", "trigonal_3bar_m_3rd_pos",
    "trigonal_3bar", "tetragonal_class_4_slash_mmm", "tetragonal_class_4_slash_m",
    "orthorhombic", "monoclinic", "triclinic". The names that don't
    correspond to a crystal system are taken from
    https://dictionary.iucr.org/Laue_class

    Note that this is a novel classification as far as the authors are aware.
    Most crystallography texts on the subject will show 9 classes. This function
    returns 10 possibilities, because the structure of the components of
    the elasticity tensor for Laue class -3m depends on whether the twofold
    axis is aligned along the Cartesian x or y axis. Assuming one adopts a
    standard orientation of the hexagonal unit cell w.r.t. the Cartesian coordinate
    system, as we do, different space groups in this Laue class have different
    forms.

    See https://doi.org/10.1017/CBO9781139017657.008 pg 232 for a review
    of the subject.
    """
    _check_space_group(sgnum)
    sgnum = int(sgnum)

    if sgnum < 3:
        return "triclinic"
    elif sgnum < 16:
        return "monoclinic"
    elif sgnum < 75:
        return "orthorhombic"
    elif sgnum < 89:
        return "tetragonal_4_slash_m"
    elif sgnum < 143:
        return "tetragonal_4_slash_mmm"
    elif sgnum < 149:
        return "trigonal_3bar"
    elif sgnum < 168:
        # Determine if this is one of the groups with the 2-fold operation in the
        # third position (e.g. 149:P312), which has different equations
        if sgnum in (149, 151, 153, 157, 159, 162, 163):
            return "trigonal_3bar_m_3rd_pos"
        else:
            return "trigonal_3bar_m_2nd_pos"
    elif sgnum < 195:
        return "hexagonal"
    else:
        return "cubic"


def voigt_elast_compon_eqn(sgnum: Union[int, str]) -> Dict:
    """
    Get the algebraic equations describing the symmetry restrictions
    on the elasticity matrix in Voigt form given a space group number.
    The unit cell
    must be in the orientation defined in doi.org/10.1016/j.commatsci.2017.01.017
    for these equations to be correct.

    Returns:
        Encoding of symmetry restrictions on elasticity matrices.
        The keys are the Voigt indices of non-independent components.
        The values are a pair of lists representing the linear combination
        of the unique compoonents that is used to determine the non-unique component
        specified in the key. The first list is the coefficients, the second
        is the indices. If a non-independent component is zero, this is indicated
        by a value of None. Any components not listed as a key are assumed to
        be independent. Only the upper triangle (i<j) is listed. Indices are
        one-based.
    """
    ELASTICITY_MATRIX_EQNS = {
        "cubic": {
            (1, 3): ([1], [(1, 2)]),
            (1, 4): None,
            (1, 5): None,
            (1, 6): None,
            (2, 2): ([1], [(1, 1)]),
            (2, 3): ([1], [(1, 2)]),
            (2, 4): None,
            (2, 5): None,
            (2, 6): None,
            (3, 3): ([1], [(1, 1)]),
            (3, 4): None,
            (3, 5): None,
            (3, 6): None,
            (4, 5): None,
            (4, 6): None,
            (5, 5): ([1], [(4, 4)]),
            (5, 6): None,
            (6, 6): ([1], [(4, 4)]),
        },
        "hexagonal": {
            (1, 4): None,
            (1, 5): None,
            (1, 6): None,
            (2, 2): ([1], [(1, 1)]),
            (2, 3): ([1], [(1, 3)]),
            (2, 4): None,
            (2, 5): None,
            (2, 6): None,
            (3, 4): None,
            (3, 5): None,
            (3, 6): None,
            (4, 5): None,
            (4, 6): None,
            (5, 5): ([1], [(4, 4)]),
            (5, 6): None,
            (6, 6): ([0.5, -0.5], [(1, 1), (1, 2)]),
        },
        "trigonal_3bar_m_2nd_pos": {
            (1, 5): None,
            (1, 6): None,
            (2, 2): ([1], [(1, 1)]),
            (2, 3): ([1], [(1, 3)]),
            (2, 4): ([-1], [(1, 4)]),
            (2, 5): None,
            (2, 6): None,
            (3, 4): None,
            (3, 5): None,
            (3, 6): None,
            (4, 5): None,
            (4, 6): None,
            (5, 5): ([1], [(4, 4)]),
            (5, 6): ([1], [(1, 4)]),
            (6, 6): ([0.5, -0.5], [(1, 1), (1, 2)]),
        },
        "trigonal_3bar_m_3rd_pos": {
            (1, 4): None,
            (1, 6): None,
            (2, 2): ([1], [(1, 1)]),
            (2, 3): ([1], [(1, 3)]),
            (2, 4): None,
            (2, 5): ([-1], [(1, 5)]),
            (2, 6): None,
            (3, 4): None,
            (3, 5): None,
            (3, 6): None,
            (4, 5): None,
            (4, 6): ([-1], [(1, 5)]),
            (5, 5): ([1], [(4, 4)]),
            (5, 6): None,
            (6, 6): ([0.5, -0.5], [(1, 1), (1, 2)]),
        },
        "trigonal_3bar": {
            (1, 6): None,
            (2, 2): ([1], [(1, 1)]),
            (2, 3): ([1], [(1, 3)]),
            (2, 4): ([-1], [(1, 4)]),
            (2, 5): ([-1], [(1, 5)]),
            (2, 6): None,
            (3, 4): None,
            (3, 5): None,
            (3, 6): None,
            (4, 5): None,
            (4, 6): ([-1], [(1, 5)]),
            (5, 5): ([1], [(4, 4)]),
            (5, 6): ([1], [(1, 4)]),
            (6, 6): ([0.5, -0.5], [(1, 1), (1, 2)]),
        },
        "tetragonal_4_slash_mmm": {
            (1, 4): None,
            (1, 5): None,
            (1, 6): None,
            (2, 2): ([1], [(1, 1)]),
            (2, 3): ([1], [(1, 3)]),
            (2, 4): None,
            (2, 5): None,
            (2, 6): None,
            (3, 4): None,
            (3, 5): None,
            (3, 6): None,
            (4, 5): None,
            (4, 6): None,
            (5, 5): ([1], [(4, 4)]),
            (5, 6): None,
        },
        "tetragonal_4_slash_m": {
            (1, 4): None,
            (1, 5): None,
            (2, 2): ([1], [(1, 1)]),
            (2, 3): ([1], [(1, 3)]),
            (2, 4): None,
            (2, 5): None,
            (2, 6): ([-1], [(1, 6)]),
            (3, 4): None,
            (3, 5): None,
            (3, 6): None,
            (4, 5): None,
            (4, 6): None,
            (5, 5): ([1], [(4, 4)]),
            (5, 6): None,
        },
        "orthorhombic": {
            (1, 4): None,
            (1, 5): None,
            (1, 6): None,
            (2, 4): None,
            (2, 5): None,
            (2, 6): None,
            (3, 4): None,
            (3, 5): None,
            (3, 6): None,
            (4, 5): None,
            (4, 6): None,
            (5, 6): None,
        },
        "monoclinic": {
            (1, 4): None,
            (1, 6): None,
            (2, 4): None,
            (2, 6): None,
            (3, 4): None,
            (3, 6): None,
            (4, 5): None,
            (5, 6): None,
        },
        "triclinic": {},
    }

    # error check typing in the above dicts
    for eqn in ELASTICITY_MATRIX_EQNS.values():
        # only unique keys
        assert sorted(list(set(eqn.keys()))) == sorted(list(eqn.keys()))
        # check that all components appearing in RHS of relations are independent, i.e.
        # they don't appear as a key
        for dependent_component in eqn:
            if eqn[dependent_component] is not None:
                for independent_component in eqn[dependent_component][1]:
                    assert not (independent_component in eqn)

    return ELASTICITY_MATRIX_EQNS[voigt_elast_class(sgnum)]


def voigt_elast_struct_svg(sgnum: Union[int, str], dest_filename: str) -> None:
    """
    Write a copy of the image showing the structure of the Voigt elasticity matrix for
    the specified space group
    """
    src_filename = os.path.join(DATA_DIR, "elast_" + voigt_elast_class(sgnum) + ".svg")
    shutil.copyfile(src_filename, dest_filename)


def indep_elast_compon_names_and_values_from_voigt(
    voigt: npt.ArrayLike, sgnum: Union[int, str]
) -> Tuple[List[str], List[float]]:
    """
    From an elasticity matrix in Voigt order and a space group number,
    extract the elastic constants that should be unique (cij where first i is as low as
    possible, then j)
    """
    eqn = voigt_elast_compon_eqn(sgnum)

    elastic_constants_names = []
    elastic_constants_values = []

    # first, figure out which constants are unique and extract them
    for i in range(1, 7):
        for j in range(i, 7):
            if (i, j) not in eqn:
                elastic_constants_names.append("c" + str(i) + str(j))
                elastic_constants_values.append(voigt[i - 1, j - 1])

    return elastic_constants_names, elastic_constants_values


def calc_bulk(elastic_constants):
    """
    Compute the bulk modulus given the elastic constants matrix in
    Voigt ordering.

    Parameters:
        elastic_constants : float
            A 6x6 numpy array containing the elastic constants in
            Voigt ordering. The material can have arbitrary anisotropy.

    Returns:
        bulk : float
            The bulk modulus, defined as the ratio between the hydrostatic
            stress (negative of the pressure p) in hydrostatic loading and
            the diltation e (trace of the strain tensor), i.e. B = -p/e
    """
    # Compute bulk modulus, based on exercise 6.14 in Tadmor, Miller, Elliott,
    # Continuum Mechanics and Thermodynamics, Cambridge University Press, 2012.
    rank_elastic_constants = np.linalg.matrix_rank(elastic_constants)
    elastic_constants_aug = np.concatenate(
        (elastic_constants, np.transpose([[1, 1, 1, 0, 0, 0]])), 1
    )
    rank_elastic_constants_aug = np.linalg.matrix_rank(elastic_constants_aug)
    if rank_elastic_constants_aug > rank_elastic_constants:
        assert rank_elastic_constants_aug == rank_elastic_constants + 1
        logger.info(
            "Information: Hydrostatic pressure not in the image of the elasticity "
            "matrix, zero bulk modulus!"
        )
        return 0.0
    else:
        # if a solution exists for a stress state of [1,1,1,0,0,0],
        # you can always use the pseudoinverse
        compliance = np.linalg.pinv(elastic_constants)
        bulk = 1 / np.sum(compliance[0:3, 0:3])
    return bulk


def map_to_Kelvin(C: npt.ArrayLike) -> npt.ArrayLike:
    """
    Compute the Kelvin form of the input 6x6 Voigt matrix
    """
    Ch = C.copy()
    Ch[0:3, 3:6] *= math.sqrt(2.0)
    Ch[3:6, 0:3] *= math.sqrt(2.0)
    Ch[3:6, 3:6] *= 2.0
    return Ch


def function_of_matrix(A, f):
    """Compute the function of a matrix"""
    ev, R = np.linalg.eigh(A)
    Dtilde = np.diag([f(e) for e in ev])
    return np.matmul(np.matmul(R, Dtilde), np.transpose(R))


def find_nearest_isotropy(elastic_constants):
    """
    Compute the distance between the provided matrix of elastic constants
    in Voigt notation, to the nearest matrix of elastic constants for an
    isotropic material. Return this distance, and the isotropic bulk and
    shear modulus.

    Ref: Morin, L; Gilormini, P and Derrien, K,
         "Generalized Euclidean Distances for Elasticity Tensors",
         Journal of Elasticity, Vol 138, pp. 221-232 (2020).

    Parameters:
        elastic_constants : float
            A 6x6 numpy array containing the elastic constants in
            Voigt ordering. The material can have arbitrary anisotropy.

    Returns:
        d : float
            Distance to the nearest elastic constants.
            log Euclidean metric.
        kappa : float
            Isotropic bulk modulus
        mu : float
            Isotropic shear modulus
    """
    E0 = 1.0  # arbitrary scaling constant (result unaffected by it)

    JJ = np.zeros(shape=(6, 6))
    KK = np.zeros(shape=(6, 6))
    v = {0: [0, 0], 1: [1, 1], 2: [2, 2], 3: [1, 2], 4: [0, 2], 5: [0, 1]}
    for ii in range(6):
        for jj in range(6):
            # i j k l = v[ii][0] v[ii][1] v[jj][0] v[jj][1]
            JJ[ii][jj] = (1.0 / 3.0) * (v[ii][0] == v[ii][1]) * (v[jj][0] == v[jj][1])
            KK[ii][jj] = (1.0 / 2.0) * (
                (v[ii][0] == v[jj][0]) * (v[ii][1] == v[jj][1])
                + (v[ii][0] == v[jj][1]) * (v[ii][1] == v[jj][0])
            ) - JJ[ii][jj]
    Chat = map_to_Kelvin(elastic_constants)
    JJhat = map_to_Kelvin(JJ)
    KKhat = map_to_Kelvin(KK)

    # Eqn (49) in Morin et al.
    fCoverE0 = function_of_matrix(Chat / E0, math.log)
    kappa = (E0 / 3.0) * math.exp(np.einsum("ij,ij", fCoverE0, JJhat))
    mu = (E0 / 2.0) * math.exp(0.2 * np.einsum("ij,ij", fCoverE0, KKhat))

    # Eqn (47) in Morin et al.
    dmat = (
        fCoverE0 - math.log(3.0 * kappa / E0) * JJhat - math.log(2.0 * mu / E0) * KKhat
    )
    d = math.sqrt(np.einsum("ij,ij", dmat, dmat))

    # Return results
    return d, kappa, mu



#!/usr/bin/python

import math

import numpy as np
from ase.build import bulk
from ase.calculators.kim.kim import KIM
from ase.geometry import get_duplicate_atoms
from ase.io import read
from elastic_excerpt import algebraically_reconstruct_matrix
from sympy import matrix2numpy, symbols

from kim_tools import (
    CENTERING_DIVISORS,
    change_of_basis_atoms,
    get_atoms_from_crystal_structure,
    get_change_of_basis_matrix_to_conventional_cell_from_formal_bravais_lattice,
    get_crystal_structure_from_atoms,
    get_formal_bravais_lattice_from_space_group,
    get_primitive_genpos_ops,
    get_space_group_number_from_prototype,
    get_symbolic_cell_from_formal_bravais_lattice,
)

# Disorganized to test it here, but the operations are really redundant
# with testing transform_atoms
from kim_tools.aflow_util.core import get_atom_indices_for_each_wyckoff_orb
from kim_tools.symmetry_util.core import (
    FixProvidedSymmetry,
    PeriodExtensionException,
    fit_voigt_tensor_and_error_to_cell_and_space_group,
    fit_voigt_tensor_to_cell_and_space_group,
    reduce_and_avg,
    transform_atoms,
)


def test_change_of_basis_atoms(
    atoms_conventional=bulk("SiC", "zincblende", 4.3596, cubic=True)
):
    calc = KIM("LennardJones612_UniversalShifted__MO_959249795837_003")
    atoms_conventional.calc = calc
    crystal_structure = get_crystal_structure_from_atoms(atoms_conventional)
    prototype_label = crystal_structure["prototype-label"]["source-value"]
    sgnum = get_space_group_number_from_prototype(prototype_label)
    formal_bravais_lattice = get_formal_bravais_lattice_from_space_group(sgnum)
    primitive_to_conventional_change_of_basis = (
        get_change_of_basis_matrix_to_conventional_cell_from_formal_bravais_lattice(
            formal_bravais_lattice
        )
    )
    conventional_to_primitive_change_of_basis = np.linalg.inv(
        primitive_to_conventional_change_of_basis
    )
    centering = formal_bravais_lattice[1]
    multiplier = np.linalg.det(primitive_to_conventional_change_of_basis)
    assert np.isclose(multiplier, CENTERING_DIVISORS[centering])
    conventional_energy = atoms_conventional.get_potential_energy()
    atoms_primitive = change_of_basis_atoms(
        atoms_conventional, conventional_to_primitive_change_of_basis
    )
    atoms_primitive.calc = calc
    primitive_energy = atoms_primitive.get_potential_energy()
    assert np.isclose(primitive_energy * multiplier, conventional_energy)
    atoms_conventional_rebuilt = change_of_basis_atoms(
        atoms_primitive, primitive_to_conventional_change_of_basis
    )
    atoms_conventional_rebuilt.calc = calc
    conventional_rebuilt_energy = atoms_conventional_rebuilt.get_potential_energy()
    assert np.isclose(conventional_energy, conventional_rebuilt_energy)


def test_test_reduced_distances():
    data_file_has_period_extension = {
        "structures/FeP_unstable.extxyz": True,
        "structures/FeP_stable.extxyz": False,
    }
    repeat = [10, 10, 10]
    for data_file in data_file_has_period_extension:
        has_period_extension = data_file_has_period_extension[data_file]
        atoms = read(data_file)
        try:
            reduce_and_avg(atoms, repeat)
            assert not has_period_extension
        except PeriodExtensionException:
            assert has_period_extension


def test_fit_voigt_tensor_to_cell_and_space_group():
    a, b, c, alpha, beta, gamma = symbols("a b c alpha beta gamma")
    # taken from A5B11CD8E_aP26_1_5a_11a_a_8a_a-001
    test_substitutions = [
        (a, 1.0),
        (b, 1.20466246551),
        (c, 1.81123604761),
        (alpha, math.radians(76.515)),
        (beta, math.radians(81.528)),
        (gamma, math.radians(71.392)),
    ]
    # Generate a random symmetric matrix
    c = np.random.random((6, 6))
    c = c + c.T
    for sgnum in (1, 3, 16, 142, 143, 152, 153, 168, 195):
        lattice = get_formal_bravais_lattice_from_space_group(sgnum)
        symbolic_cell = get_symbolic_cell_from_formal_bravais_lattice(lattice)
        cell = matrix2numpy(symbolic_cell.subs(test_substitutions), dtype=float)
        # Throw some exponents in there
        cell += np.random.rand(3, 3) * 1e-16

        # Not checking the error here, so just pass c itself as a dummy error
        c_mat_symm_rot_sympy, _ = fit_voigt_tensor_and_error_to_cell_and_space_group(
            c, c, cell, sgnum
        )
        c_mat_symm_rot = fit_voigt_tensor_to_cell_and_space_group(c, cell, sgnum)
        if lattice == "aP":
            assert np.allclose(c, c_mat_symm_rot_sympy)
            assert np.allclose(c, c_mat_symm_rot)
        # This takes any matrix, picks out the unique constants based on the
        # algebraic diagrams, and returns a matrix conforming to the material symmetry
        c_mat_symm_alg = algebraically_reconstruct_matrix(c_mat_symm_rot_sympy, sgnum)
        assert np.allclose(c_mat_symm_rot_sympy, c_mat_symm_alg)
        assert np.allclose(c_mat_symm_rot, c_mat_symm_alg)

    # Use SG 75 to test a specific matrix including error
    sgnum = 75
    lattice = get_formal_bravais_lattice_from_space_group(sgnum)
    symbolic_cell = get_symbolic_cell_from_formal_bravais_lattice(lattice)
    cell = matrix2numpy(symbolic_cell.subs(test_substitutions), dtype=float)
    # Throw some exponents in there
    cell += np.random.rand(3, 3) * 1e-16

    c = np.ones((6, 6))
    c_err = np.ones((6, 6))

    # In Laue Class 4/m, [0,5]=-[1,5].
    # Here we have [0,5]=0, [1,5]=1, so
    # we should end up with [0,5]=-1/2,
    # [1,5]=1/2, and the resulting variance
    # be 1/4 of the sum of their variances
    c[0, 5] = 0
    c[5, 0] = 0
    c_err[1, 5] = 15 ** (1 / 2)
    c_err[5, 1] = 15 ** (1 / 2)

    ref_out = (
        [
            [1.0, 1.0, 1.0, 0.0, 0.0, -0.5],
            [1.0, 1.0, 1.0, 0.0, 0.0, 0.5],
            [1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            [-0.5, 0.5, 0.0, 0.0, 0.0, 1.0],
        ],
        [
            [0.5**0.5, 0.5**0.5, 0.5**0.5, 0.0, 0.0, 2.0],
            [0.5**0.5, 0.5**0.5, 0.5**0.5, 0.0, 0.0, 2.0],
            [0.5**0.5, 0.5**0.5, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.5**0.5, 0.5**0.5, 0.0],
            [0.0, 0.0, 0.0, 0.5**0.5, 0.5**0.5, 0.0],
            [2.0, 2.0, 0.0, 0.0, 0.0, 1.0],
        ],
    )
    ref_out_symm = (
        [
            [1.0, 1.0, 1.0, 0.0, 0.0, -0.5],
            [1.0, 1.0, 1.0, 0.0, 0.0, 0.5],
            [1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            [-0.5, 0.5, 0.0, 0.0, 0.0, 1.0],
        ],
        [
            [0.5**0.5, 1, 0.5**0.5, 0.0, 0.0, 2.0],
            [1, 0.5**0.5, 0.5**0.5, 0.0, 0.0, 2.0],
            [0.5**0.5, 0.5**0.5, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.5**0.5, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.5**0.5, 0.0],
            [2.0, 2.0, 0.0, 0.0, 0.0, 1.0],
        ],
    )
    assert np.allclose(
        fit_voigt_tensor_and_error_to_cell_and_space_group(c, c_err, cell, sgnum),
        ref_out,
    )
    assert np.allclose(
        fit_voigt_tensor_and_error_to_cell_and_space_group(c, c_err, cell, sgnum, True),
        ref_out_symm,
    )
    assert np.allclose(
        fit_voigt_tensor_to_cell_and_space_group(c, cell, sgnum),
        ref_out[0],
    )


def test_FixProvidedSymmetry():
    test_cases = [
        {
            "symm": 1,
            "deform_allowed": True,
        },
        {
            "symm": get_primitive_genpos_ops(1),
            "deform_allowed": True,
        },
        {
            "symm": 221,
            "deform_allowed": False,
        },
        {
            "symm": get_primitive_genpos_ops(221),
            "deform_allowed": False,
        },
    ]
    for case in test_cases:
        atoms = bulk("CsCl", "cesiumchloride", 1.0)
        constraint = FixProvidedSymmetry(atoms, case["symm"])
        atoms.set_constraint(constraint)
        atoms.set_positions(
            [
                [
                    0,
                    0,
                    0,
                ],
                [0.6, 0.5, 0.5],
            ]
        )
        deform_allowed = case["deform_allowed"]
        assert np.allclose(atoms[1].position, [0.6, 0.5, 0.5]) == deform_allowed
        new_cell = [[1.1, 0, 0], [0, 1, 0], [0, 0, 1]]
        atoms.set_cell(new_cell, scale_atoms=True)
        assert np.allclose(atoms.cell, new_cell) == deform_allowed


def test_transform_atoms_and_symmetry_and_wyckoff_indices(input_crystal_structures):
    for material in input_crystal_structures:
        atoms = get_atoms_from_crystal_structure(material)
        prototype_label = material["prototype-label"]["source-value"]
        wyckoff_atom_indices = get_atom_indices_for_each_wyckoff_orb(prototype_label)
        sgnum = get_space_group_number_from_prototype(prototype_label)
        genpos_ops = get_primitive_genpos_ops(sgnum)
        # Check that transforming by a symmetry op always gets an identical
        # atoms object
        for op in genpos_ops:
            atoms_transformed = transform_atoms(atoms, op)
            atoms_combined = atoms_transformed + atoms
            dupes = get_duplicate_atoms(atoms_combined, delete=True)
            assert len(atoms_combined) == len(atoms)
            # Check that any duplicates correspond to the same Wyckoff orbit
            for dupe in dupes:
                for position in wyckoff_atom_indices:
                    if dupe[0] in position["indices"]:
                        assert dupe[1] - len(atoms) in position["indices"]


if __name__ == "__main__":
    test_fit_voigt_tensor_to_cell_and_space_group()



#!/usr/bin/python

import os
import subprocess

import kim_edn
import numpy as np
import numpy.typing as npt
from ase.atoms import Atoms
from ase.build import bulk
from ase.calculators.lj import LennardJones

from kim_tools import (
    KIMTestDriver,
    SingleCrystalTestDriver,
    cartesian_rotation_is_in_point_group,
    detect_unique_crystal_structures,
    get_deduplicated_property_instances,
)
from kim_tools.test_driver.core import _get_optional_source_value


class TestInitSingleCrystalTestDriver(SingleCrystalTestDriver):
    def _calculate(self, **kwargs) -> None:
        pass


class TestInitKIMTestDriver(KIMTestDriver):
    def _calculate(self, **kwargs) -> None:
        pass


class TestIsolatedEnergyDriver(KIMTestDriver):
    def _calculate(self, species):
        """
        Example calculate method for testing isolated energy getter
        """
        assert np.isclose(self.get_isolated_energy_per_atom(species), 0.0)


class TestTestDriver(KIMTestDriver):
    def _calculate(self, property_name, species):
        """
        example calculate method

        Args:
            property_name: for testing ability to find properties at different paths.
            !!! AN ACTUAL TEST DRIVER SHOULD NOT HAVE AN ARGUMENT SUCH AS THIS !!!
        """
        atoms = Atoms([species], [[0, 0, 0]])
        self._add_property_instance(property_name, "This is an example disclaimer.")
        self._add_key_to_current_property_instance(
            "species", atoms.get_chemical_symbols()[0]
        )
        self._add_key_to_current_property_instance(
            "mass", atoms.get_masses()[0], "amu", {"source-std-uncert-value": 1}
        )


class TestStructureDetectionTestDriver(SingleCrystalTestDriver):
    def _calculate(self, deform_matrix: npt.NDArray = np.eye(3), **kwargs):
        """
        strain the crystal and write a crystal-structure-npt.
        For testing various crystal detection things
        """
        atoms = self._get_atoms()
        original_cell = atoms.cell
        new_cell = original_cell @ deform_matrix
        atoms.set_cell(new_cell, scale_atoms=True)
        self._update_nominal_parameter_values(atoms)
        self._add_property_instance_and_common_crystal_genome_keys(
            "crystal-structure-npt"
        )


class FileWritingTestDriver(KIMTestDriver):
    def _calculate(self):
        """
        Mock calculate for file writing testing
        """
        self._add_property_instance("file-prop")
        with open("foo.txt", "w") as f:
            f.write("foo")
        self._add_file_to_current_property_instance("textfile", "foo.txt")


def test_kimtest(monkeypatch):
    test = TestTestDriver(LennardJones())
    testing_property_names = [
        "atomic-mass",  # already in kim-properties
        "atomic-mass0",  # found in $PWD/local-props
        "atomic-mass0",  # check that repeat works fine
        # check that full id works as well, found in $PWD/local-props
        "tag:brunnels@noreply.openkim.org,2016-05-11:property/atomic-mass1",
        "atomic-mass2",  # found in $PWD/local-props/atomic-mass2
        "atomic-mass3",  # found in $PWD/mock-test-drivers-dir/mock-td/local_props,
        # tested using the monkeypatch below
    ]

    monkeypatch.setenv(
        "KIM_PROPERTY_PATH",
        os.path.join(os.getcwd(), "mock-test-drivers-dir/*/local-props")
        + ":"
        + os.path.join(os.getcwd(), "mock-test-drivers-dir/*/local_props"),
    )

    for prop_name in testing_property_names:
        test(property_name=prop_name, species="Ar")

    assert len(test.property_instances) == 6
    test.write_property_instances_to_file()


def test_detect_unique_crystal_structures():
    reference_structure = kim_edn.load("structures/OSi.edn")
    test_structure = kim_edn.load("structures/OSi_twin.edn")
    assert (
        len(
            detect_unique_crystal_structures(
                [
                    reference_structure,
                    reference_structure,
                    test_structure,
                    test_structure,
                    test_structure,
                    test_structure,
                ],
                allow_rotation=True,
            )
        )
        == 1
    )
    assert (
        len(
            detect_unique_crystal_structures(
                [
                    reference_structure,
                    reference_structure,
                    test_structure,
                    test_structure,
                    test_structure,
                    test_structure,
                ],
                allow_rotation=False,
            )
        )
        == 2
    )


def test_get_deduplicated_property_instances():
    property_instances = kim_edn.load("structures/results.edn")
    fully_deduplicated = get_deduplicated_property_instances(property_instances)
    assert len(fully_deduplicated) == 6
    inst_with_1_source = 0
    inst_with_2_source = 0
    for property_instance in fully_deduplicated:
        n_inst = len(
            property_instance["crystal-genome-source-structure-id"]["source-value"][0]
        )
        if n_inst == 1:
            inst_with_1_source += 1
        elif n_inst == 2:
            inst_with_2_source += 1
        else:
            assert False
    assert inst_with_1_source == 3
    assert inst_with_2_source == 3
    partially_deduplicated = get_deduplicated_property_instances(
        property_instances, ["mass-density-crystal-npt"]
    )
    assert len(partially_deduplicated) == 8
    inst_with_1_source = 0
    inst_with_2_source = 0
    for property_instance in partially_deduplicated:
        n_inst = len(
            property_instance["crystal-genome-source-structure-id"]["source-value"][0]
        )
        if n_inst == 1:
            inst_with_1_source += 1
        elif n_inst == 2:
            inst_with_2_source += 1
        else:
            assert False
    assert inst_with_1_source == 7
    assert inst_with_2_source == 1


def test_structure_detection():
    test = TestStructureDetectionTestDriver(LennardJones())
    atoms = bulk("Mg")
    hcp_prototype = "A_hP2_194_c"
    hcp_library_prototype = "A_hP2_194_c-001"
    hcp_shortname = ["Hexagonal Close Packed (Mg, $A3$, hcp) Structure"]
    stretch = np.diag([1, 1, 10])
    for deform, prototype, library_prototype, shortname in zip(
        [np.eye(3), stretch],
        [hcp_prototype, hcp_prototype],
        [hcp_library_prototype, None],
        [hcp_shortname, None],
    ):
        property_instance = test(atoms, deform_matrix=deform)[0]
        assert property_instance["prototype-label"]["source-value"] == prototype
        assert (
            _get_optional_source_value(property_instance, "library-prototype-label")
            == library_prototype
        )
        assert _get_optional_source_value(property_instance, "short-name") == shortname


def test_get_isolated_energy_per_atom():
    for model in [
        LennardJones(),
        "LennardJones612_UniversalShifted__MO_959249795837_003",
        "Sim_LAMMPS_LJcut_AkersonElliott_Alchemy_PbAu",
    ]:
        td = TestIsolatedEnergyDriver(model)
        for species in ["Pb", "Au"]:
            td(species=species)


def test_init_rotation():
    td = TestInitSingleCrystalTestDriver(LennardJones())
    atoms = bulk("Mg", crystalstructure="bct", a=1, c=3.14)
    atoms.rotate(60, "z", rotate_cell=True)
    # We rotated from standard orientation by 60 deg
    # ccw. So when the atoms get rebuilt, the rotation
    # we will do to standardize is 60 deg cw.
    ref_rotation = np.array(
        [  # 60 deg clockwise
            [0.5, 0.866025, 0.0],
            [-0.866025, 0.5, 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    td(atoms)
    input_rotation = td.get_input_rotation()
    assert cartesian_rotation_is_in_point_group(
        ref_rotation.T @ input_rotation,
        139,
        atoms.cell,
    )


def test_file_writing():
    td = FileWritingTestDriver(LennardJones())
    td()
    td.write_property_instances_to_file()
    assert os.path.isfile("output/foo-1.txt")
    # add a file into output to make sure it's not being moved
    with open("output/bar", "w") as f:
        f.write("bar")
    td.write_property_instances_to_file("foo.edn")
    # will raise a filenotfound error if missing
    os.remove("foo.edn")
    os.remove("foo-1.txt")
    os.remove("output/bar")


def test_atom_style():
    td = TestInitKIMTestDriver("LennardJones612_UniversalShifted__MO_959249795837_003")
    assert td._get_supported_lammps_atom_style() == "atomic"
    charge_sm = "Sim_LAMMPS_ReaxFF_AnGoddard_2015_BC__SM_389039364091_000"
    subprocess.run(
        f"kim-api-collections-management install --force CWD {charge_sm}",
        shell=True,
        check=True,
    )
    td = TestInitKIMTestDriver(charge_sm)
    assert td._get_supported_lammps_atom_style() == "charge"
    subprocess.run(
        f"kim-api-collections-management remove --force {charge_sm}",
        shell=True,
        check=True,
    )


if __name__ == "__main__":
    test_atom_style()
